/* This header allows you to easily use TI-BASIC as an extension language.

   Copyright (C) Kevin Kofler, 2004. All rights reserved.

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included
   in all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
   OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
   DEALINGS IN THE SOFTWARE. */

typedef void (* APP_EXT_FUNC)(void);
typedef struct SAppExtension
{
   unsigned long name;
   unsigned long help;
   unsigned short index;
} APP_EXTENSION;
typedef struct SAppExtEntry
{
   APP_EXT_FUNC extension;
   unsigned short flags;
} APP_EXT_ENTRY;
enum {APP_EXT_PROGRAM=0x0000, APP_EXT_FUNCTION=0x0001};
#define OO_FIRST_STRING             2048
#define OO_FIRST_APP_STRING         2048
#define OO_APPSTRING (OO_FIRST_STRING+OO_FIRST_APP_STRING)

#define OO_HANDLE (0xFF000000u)
#define OO_SYSTEM_FRAME OO_HANDLE
typedef enum {OO_RW=0,  OO_RO=1,
              OO_SEQ=0, OO_KEYED=2} OO_Flags;
typedef struct
{
   ULONG key;
   void *value;
} OO_Attr;
typedef struct SFrameHdr
{
   pFrame parent;
   pFrame prototype;
   OO_Flags flags;
   ULONG first;
   ULONG count;
} OO_Hdr;
typedef struct SFrame
{
   OO_Hdr head;
   OO_Attr pair[];
} Frame;
typedef void (* const OO_MethodPtr)(void);
#define FRAME(name, parent, proto, first, count) \
const Frame name = {  \
{                     \
   (pFrame)parent,    \
   (pFrame)proto,     \
   OO_RO | OO_KEYED,  \
   first,             \
   count              \
},                    \
{
#define ATTR(selector, val) {selector, (void *)(val)},
#define STRING_ATTR(sel, s) {OO_FIRST_STRING+(sel), s},
#define ENDFRAME }};
typedef enum {APP_NONE=0,
              APP_INTERACTIVE=1,
              APP_CON=2,
              APP_ACCESS_SYSVARS=4,
              APP_BACKGROUND=8}
        APP_Flags;

#define EXTENSIONS APP_EXTENSION const extensions[] = {
#define DEFINE_EXTENSION(ext) {OO_APPSTRING+((ext)<<1), OO_APPSTRING+((ext)<<1)+1, (ext)},
#define ENDEXTENSIONS };

#define EXT_ENTRIES APP_EXT_ENTRY const extEntries[] = {
#define DEFINE_EXT_ENTRY(extptr,exttype) {(extptr), (exttype)},
#define ENDEXT_ENTRIES };

#define BASEXTFRAME(appname, extcount) \
  FRAME(tibasicextFrame, OO_SYSTEM_FRAME, 0, OO_APP_FLAGS, (((extcount)<<1)+7)) \
    ATTR(OO_APP_FLAGS, APP_NONE) \
    ATTR(OO_APP_NAME, (appname)) \
    ATTR(OO_APP_TOK_NAME, (appname)) \
    ATTR(OO_APP_EXT_COUNT, (extcount)) \
    ATTR(OO_APP_EXTENSIONS, extensions) \
    ATTR(OO_APP_EXT_ENTRIES, extEntries) \
    ATTR(OO_APP_ABOUT, NoCallBack)
#define DEFINE_EXT_STRINGS(ext, extname, extdesc) \
  ATTR(OO_APPSTRING+((ext)<<1), (extname)) \
  ATTR(OO_APPSTRING+((ext)<<1)+1, (extdesc))

pFrame saveframe;
#define InstallBasExts() (OO_InstallAppHookByName ("TIHOME", (pFrame)&tibasicextFrame,&saveframe))
#define UninstallBasExts() (OO_UninstallAppHookByName ("TIHOME", saveframe))

static inline void progrun(const char *name)
{
  size_t l=strlen(name);
  HANDLE h=HeapAlloc(l+7);
  ESI p=HeapDeref(h);
  *(((short*)p)++)=l+5;
  *(((short*)p)++)=(ENDSTACK_TAG<<8)+END_TAG;
  *(p++)=0;
  strcpy(p,name);
  p+=l+1;
  *p=USERFUNC_TAG;
  TRY
    NG_execute (h, FALSE);
  FINALLY
    HeapFree (h);
  ENDFINAL
}
